@@ -22,6 +22,7 @@
 #include <gst/play/gstplay.h>
 #include <gst/play/gstplay-visualization.h>
 #include <gst/play/gstplay-signal-adapter.h>
+#include <gst/play/gstplay-video-overlay-video-renderer.h>
 
 #include <adwaita.h>
 #include <glib/gi18n.h>
@@ -32,6 +33,8 @@ enum {
   PROP_PLAYBACK_SPEED,
   PROP_STATE,
   PROP_TITLE,
+  PROP_DURATION,
+  PROP_POSITION,
   LAST_PROP,
 };
 static GParamSpec *props[LAST_PROP];
@@ -79,6 +82,7 @@ struct _LiviWindow
 
   GstPlay              *player;
   GstPlaySignalAdapter *signal_adapter;
+  GstElement           *gtk4paintablesink;
   GstPlayState          state;
   guint                 cookie;
 
@@ -90,6 +94,8 @@ struct _LiviWindow
     char               *title;
     char               *ref_uri;
     gboolean            uri_preprocessed;
+    GstClockTime        duration_ns;
+    GstClockTime        position_ns;
   } stream;
 
   GtkFileFilter        *video_filter;
@@ -102,6 +108,7 @@ struct _LiviWindow
   LiviRecentVideos     *recent_videos;
 
   gboolean              have_pointer;
+  GSettings            *settings;
 };
 
 G_DEFINE_TYPE (LiviWindow, livi_window, ADW_TYPE_APPLICATION_WINDOW)
@@ -189,10 +196,15 @@ reset_stream (LiviWindow *self)
   if (self->stream.title) {
     g_free (self->stream.title);
     g_free (self->stream.ref_uri);
-    g_object_notify_by_pspec (G_OBJECT (self), props[PROP_TITLE]);
   }
   memset (&self->stream, 0, sizeof (self->stream));
   self->stream.playback_speed = 100;
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_TITLE]);
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_DURATION]);
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_POSITION]);
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_PLAYBACK_SPEED]);
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_MUTED]);
 }
 
 
@@ -260,6 +272,12 @@ livi_window_get_property (GObject    *object,
     case PROP_TITLE:
       g_value_set_string (value, self->stream.title);
       break;
+    case PROP_DURATION:
+      g_value_set_uint64 (value, self->stream.duration_ns);
+      break;
+    case PROP_POSITION:
+      g_value_set_uint64 (value, self->stream.position_ns);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -727,6 +745,9 @@ on_player_duration_changed (GstPlaySignalAdapter *adapter, guint64 duration, gpo
 
   g_debug ("Duration %" G_GUINT64_FORMAT "s", duration / GST_SECOND);
   livi_controls_set_duration (self->controls, duration);
+
+  self->stream.duration_ns = duration;
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_DURATION]);
 }
 
 
@@ -739,6 +760,12 @@ on_player_position_updated (GstPlaySignalAdapter *adapter, guint64 position, gpo
 
   g_assert (LIVI_IS_WINDOW (self));
   livi_controls_set_position (self->controls, position);
+
+  if (self->stream.position_ns == position)
+    return;
+
+  self->stream.position_ns = position;
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_POSITION]);
 }
 
 
@@ -907,6 +934,7 @@ update_video_streams (LiviWindow *self, GstPlayMediaInfo *info)
   GstPlayVisualization **vis = NULL;
   gboolean success;
   const char *visname = NULL;
+  g_autofree char *audio_vis = NULL;
 
   num_video_streams = gst_play_media_info_get_number_of_video_streams (info);
   if (num_video_streams) {
@@ -921,12 +949,17 @@ update_video_streams (LiviWindow *self, GstPlayMediaInfo *info)
     goto out;
   }
 
+  audio_vis = g_settings_get_string (self->settings, "audio-visualization");
   for (int i = 0; vis[i]; i++) {
-    if (g_strcmp0 (vis[i]->name, "wavescope"))
-      visname = "wavescope";
+    if (g_strcmp0 (vis[i]->name, audio_vis) == 0) {
+      visname = audio_vis;
+      break;
+    }
   }
-  if (!visname)
+  if (!visname) {
     visname = vis[0]->name;
+    g_warning ("Visualizer '%s' not found using '%s'", audio_vis, visname);
+  }
 
   success = gst_play_set_visualization (self->player, visname);
   if (!success) {
@@ -976,11 +1009,44 @@ on_realize (LiviWindow *self)
 
   g_assert (LIVI_IS_WINDOW (self));
 
-  surface = gtk_native_get_surface (gtk_widget_get_native (GTK_WIDGET (self->picture_video)));
-  livi_gst_paintable_realize (LIVI_GST_PAINTABLE (self->paintable), surface);
+  if (!self->paintable) {
+    g_assert (self->gtk4paintablesink != NULL);
+    /* Retrieving the paintable checks the GdkGLContext and queries the
+     * supported dmabuf formats, so only do it here during realization the
+     * first time. */
+    g_object_get (self->gtk4paintablesink, "paintable", &self->paintable, NULL);
+    g_assert (self->paintable != NULL);
+
+    gtk_picture_set_paintable (self->picture_video, self->paintable);
+  }
+
+  if (LIVI_IS_GST_PAINTABLE (self->paintable)) {
+    surface = gtk_native_get_surface (gtk_widget_get_native (GTK_WIDGET (self->picture_video)));
+    livi_gst_paintable_realize (LIVI_GST_PAINTABLE (self->paintable), surface);
+  }
 
   if (!self->player) {
-    self->player = gst_play_new (GST_PLAY_VIDEO_RENDERER (g_object_ref (self->paintable)));
+    GstPlayVideoRenderer *video_renderer;
+    GstStructure *config;
+
+    if (self->gtk4paintablesink) {
+      GstElement *video_sink;
+      GdkGLContext *gl_context;
+
+      g_object_get (self->paintable, "gl-context", &gl_context, NULL);
+      if (gl_context) {
+        video_sink = gst_element_factory_make ("glsinkbin", NULL);
+        g_object_set (video_sink, "sink", self->gtk4paintablesink, NULL);
+      } else {
+        video_sink = self->gtk4paintablesink;
+      }
+
+      video_renderer = gst_play_video_overlay_video_renderer_new_with_sink (NULL, video_sink);
+    } else {
+      video_renderer = GST_PLAY_VIDEO_RENDERER (g_object_ref (self->paintable));
+    }
+
+    self->player = gst_play_new (video_renderer);
     self->signal_adapter = gst_play_signal_adapter_new (self->player);
     g_object_connect (self->signal_adapter,
                       "signal::error", G_CALLBACK (on_player_error), self,
@@ -993,6 +1059,11 @@ on_realize (LiviWindow *self)
                       "signal::media-info-updated", G_CALLBACK (on_media_info_updated), self,
                       "signal::end-of-stream", G_CALLBACK (on_end_of_stream), self,
                       NULL);
+
+    config = gst_play_get_config (self->player);
+    /* Update position once a second (default is 100ms) */
+    gst_play_config_set_position_update_interval (config, 1000);
+    gst_play_set_config (self->player, config);
   }
 }
 
@@ -1001,12 +1072,15 @@ static gboolean
 livi_window_close_request (GtkWindow *window)
 {
   LiviWindow *self = LIVI_WINDOW (window);
+  GstClockTime pos = gst_play_get_position (self->player);
 
-  if (self->stream.ref_uri) {
+  /* Only update pos > 0 so we don't lose the current pos if
+   * e.g. resuming fails due to a network error */
+  if (self->stream.ref_uri && pos > 0) {
     livi_recent_videos_update (self->recent_videos,
                                self->stream.ref_uri,
                                self->stream.uri_preprocessed,
-                               gst_play_get_position (self->player));
+                               pos);
   }
 
   return GTK_WINDOW_CLASS (livi_window_parent_class)->close_request (window);
@@ -1021,6 +1095,7 @@ livi_window_dispose (GObject *obj)
   g_clear_pointer (&self->last_local_uri, g_free);
   g_clear_object (&self->recent_videos);
   g_clear_object (&self->signal_adapter);
+  g_clear_object (&self->gtk4paintablesink);
   g_clear_object (&self->player);
   if (self->cookie) {
     GApplication *app = g_application_get_default ();
@@ -1029,6 +1104,7 @@ livi_window_dispose (GObject *obj)
     self->cookie = 0;
   }
   reset_stream (self);
+  g_clear_object (&self->settings);
 
   G_OBJECT_CLASS (livi_window_parent_class)->dispose (obj);
 }
@@ -1070,6 +1146,16 @@ livi_window_class_init (LiviWindowClass *klass)
                          NULL,
                          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
 
+  props[PROP_DURATION] =
+    g_param_spec_uint64 ("duration", "", "",
+                         0, G_MAXUINT64, 0,
+                         G_PARAM_READABLE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
+  props[PROP_POSITION] =
+    g_param_spec_uint64 ("position", "", "",
+                         0, G_MAXUINT64, 0,
+                         G_PARAM_READABLE | G_PARAM_STATIC_STRINGS | G_PARAM_EXPLICIT_NOTIFY);
+
   g_object_class_install_properties (object_class, LAST_PROP, props);
 
   g_type_ensure (LIVI_TYPE_CONTROLS);
@@ -1143,12 +1229,28 @@ static GActionEntry win_entries[] =
 static void
 livi_window_init (LiviWindow *self)
 {
+  g_autoptr (GstElementFactory) element_factory = NULL;
+  const char *force_builtin_sink = g_getenv ("LIVI_FORCE_BUILTIN_SINK");
+
+  self->settings = g_settings_new ("org.sigxcpu.Livi");
+
   reset_stream (self);
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
-  self->paintable = livi_gst_paintable_new ();
-  gtk_picture_set_paintable (self->picture_video, self->paintable);
+  element_factory = gst_element_factory_find ("gtk4paintablesink");
+  if (!force_builtin_sink && element_factory &&
+      gst_plugin_feature_check_version (GST_PLUGIN_FEATURE (element_factory), 0, 12, 4)) {
+    self->gtk4paintablesink = gst_element_factory_create (element_factory, NULL);
+    gst_object_ref_sink (self->gtk4paintablesink);
+
+    g_debug ("Using gtk4paintablesink");
+  } else {
+    self->paintable = livi_gst_paintable_new ();
+    gtk_picture_set_paintable (self->picture_video, self->paintable);
+
+    g_debug ("Using built in sink");
+  }
 
   add_controls_toggle (self, GTK_WIDGET (self->picture_video));
   add_controls_toggle (self, GTK_WIDGET (self->revealer_center));
@@ -1176,7 +1278,7 @@ livi_window_resume_pos (LiviWindow *self)
   if (pos > 0) {
     pos *= GST_MSECOND;
     /* Seek directly without showing any overlays */
-    g_debug ("Found video %s, resuming at %ld s", self->stream.ref_uri, pos / GST_SECOND);
+    g_debug ("Found video %s, resuming at %" G_GINT64_FORMAT "s", self->stream.ref_uri, pos / GST_SECOND);
     gst_play_seek (self->player, pos);
     show_resume_or_restart_overlay (self, TRUE);
   }
